
{*****************************************************************************
 * Name           : KNGControls                                              *
 * Author         : Lesage Stefaan                                           *
 * Last Revision  : 06/07/1999                                               *
 * Description    :                                                          *
 * History        :                                                          *
 *                                                                           *
 * Date       By              Description                                    *
 * ----       --              -----------                                    *
 * 06/07/1999 Lesage Stefaan  Initial creation of the Unit.                  *
 *****************************************************************************}

unit KNGControls;

{$I KNG.INC}

interface

uses Controls, Classes, Graphics, Messages, stdctrls, extctrls, KNGClasses, Windows;

type
  TKNGBorderSide = (bsLeft, bsTop, bsRight, bsBottom);
  TKNGBorderSides = set of TKNGBorderSide;

  TKNGBevelStyle = (bvNone, bvLowered, bvRaised);
  TKNGBorderStyle = (bsNone, bsSingle);

  TKNGBevel = class(TPersistent)
  private
    FBevelInner     : TKNGBevelStyle;
    FBevelOuter     : TKNGBevelStyle;
    FBevelWidth     : TBevelWidth;
    FBevelBorders   : TKNGBorderSides;
    FBorderStyle    : TKNGBorderStyle;
    FBorderWidth    : TBorderWidth;
    FHighLightColor : TColor;
    FShadowColor    : TColor;
    FOnChange       : TNotifyEvent;

    procedure SetBevelInner(Value: TKNGBevelStyle);
    procedure SetBevelOuter(Value: TKNGBevelStyle);
    procedure SetBevelWidth(Value: TBevelWidth);
    procedure SetBevelBorders(Value: TKNGBorderSides);
    procedure SetBorderStyle(Value: TKNGBorderStyle);
    procedure SetBorderWidth(Value: TBorderWidth);
    procedure SetHighLightColor(Value: TColor);
    procedure SetShadowColor(Value: TColor);
    function GetBevelSideSize: Integer;
  protected
  public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    function ClientRect(ControlRect:TRect):TRect;
    procedure PaintBevel(Canvas:TCanvas; var Rect:TRect; Color:TColor; Transparent: Boolean);
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  published
    property BevelInner: TKNGBevelStyle read FBevelInner write SetBevelInner;
    property BevelOuter: TKNGBevelStyle read FBevelOuter write SetBevelOuter;
    property BevelWidth: TBevelWidth read FBevelWidth write SetBevelWidth;
    property BevelBorders: TKNGBorderSides read FBevelBorders write SetBevelBorders;
    property BevelSideSize: Integer read GetBevelSideSize;
    property BorderStyle: TKNGBorderStyle read FBorderStyle write SetBorderStyle;
    property BorderWidth: TBorderWidth read FBorderWidth write SetBorderWidth;
    property HighLightColor: TColor read FHighLightColor write SetHighLightColor;
    property ShadowColor: TColor read FShadowColor write SetShadowColor;
  end;

  TKNGGraphicControl = class(TGraphicControl)
  private
    FAbout          : TKNGAboutString;
    FMouseInControl : Boolean;
    FOnMouseEnter   : TNotifyEvent;
    FOnMouseLeave   : TNotifyEvent;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
  protected
    { Methods }
    procedure MouseEnter; virtual;
    procedure MouseLeave; virtual;
    { Properties }
    property About: TKNGAboutString read FAbout write FAbout;
    property OnMouseEnter: TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave: TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
  public
    property MouseInControl: Boolean read FMouseInControl;
  end;

  TKNGCustomLabel = Class(TKNGGraphicControl)
  private
    FAlignment      : TAlignment;
    FAutoSize       : Boolean;
    FBevel          : TKNGBevel;
    FFocusControl   : TWinControl;
    FLayout         : TTextLayout;
    FShowAccelChar  : Boolean;
    FWordWrap       : Boolean;

    procedure CMDialogChar(var Message: TCMDialogChar); message CM_DIALOGCHAR;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
    procedure DoDrawText(var Rect: TRect; Flags: Word);
    function GetTransparent: Boolean;
    procedure SetAlignment(Value: TAlignment);
    procedure SetBevel(Value: TKNGBevel);
    procedure SetFocusControl(Value: TWinControl);
    procedure SetLayout(Value: TTextLayout);
    procedure SetShowAccelChar(Value: Boolean);
    procedure SetTransparent(Value: Boolean);
    procedure SetWordWrap(Value: Boolean);
  protected
    procedure AdjustBounds; virtual;
    procedure BevelChanged(Sender: TObject); virtual;
    function GetLabelText: string; virtual;
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
    procedure Paint; override;
    procedure SetAutoSize(Value: Boolean); virtual;
    property Alignment: TAlignment read FAlignment write SetAlignment
      default taLeftJustify;
    property AutoSize: Boolean read FAutoSize write SetAutoSize default True;
    property Bevel: TKNGBevel read FBevel write SetBevel;
    property FocusControl: TWinControl read FFocusControl write SetFocusControl;
    property Layout: TTextLayout read FLayout write SetLayout default tlTop;
    property ShowAccelChar: Boolean read FShowAccelChar write SetShowAccelChar default True;
    property Transparent: Boolean read GetTransparent write SetTransparent default False;
    property WordWrap: Boolean read FWordWrap write SetWordWrap default False;
    property About: TKNGAboutString read FAbout write FAbout;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Canvas;
  end;


  TKNGLabel = class(TKNGCustomLabel)
  published
    property About;
    property Align;
    property Alignment;
    property AutoSize;
    property Bevel;
    property Caption;
    property Color;
    property DragCursor;
    property DragMode;
    property Enabled;
    property FocusControl;
    property Font;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowAccelChar;
    property ShowHint;
    property Transparent;
    property Layout;
    property Visible;
    property WordWrap;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseUp;
    property OnStartDrag;
  end;

  TKNGURLKind = (ukHTTP, ukMailTo);

  TKNGURLLabel = class(TKNGCustomLabel)
  private
    FAddURLKind : Boolean;
    FURLKind    : TKNGURLKind;

    { Property Get and Set Routines }
    procedure SetAddURLKind(Value: Boolean);
    procedure SetURLKind(Value: TKNGURLKind);
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
                        X,Y : Integer); override;
    procedure MouseEnter; override;
    procedure MouseLeave; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    { Properties }
    { Inherited from TCustomLabel }
    property Align;
    property Alignment;
    property AutoSize;
    property Caption;
    property Color;
    property DragCursor;
    property DragMode;
    property Enabled;
    property FocusControl;
    property Font;
    property Layout;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowAccelChar;
    property ShowHint;
    property Transparent;
    property Visible;
    property WordWrap;
    { Inherited from TKNGCustomLabel }
    property About;
    property Bevel;
    { New in TKNGURLLabel }
    property AddURLKind: Boolean read FAddURLKind write SetAddURLKind;
    property URLKind: TKNGURLKind read FURLKind write SetURLKind;
    { Delphi 4 Properties }
    {$IFDEF KNG_D4}
    property Anchors;
    property BiDiMode;
    property Constraints;
    property DragKind;
    property ParentBiDiMode;
    {$ENDIF}
    { Events }
    { Inherited from TCustomLabel }
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
    { Inherited from TKNGCustomLabel }
    property OnMouseEnter;
    property OnMouseLeave;
    { Delphi 4 Events }
    {$IFDEF KNG_D4}
    property OnEndDock;
    property OnStartDock;
    {$ENDIF}
  end;

  TKNGCustomEdit = class(TCustomEdit)
  private
    FAbout: TKNGAboutString;
    FColorDefault: TColor;
    FColorDisabled: TColor;
    FColorFocused: TColor;
    FEnabled: Boolean;
    FExitOnReturn: Boolean;
    FFocusOnMouseEnter: Boolean;
    FFont: TFont;
    FFontColorDefault: TColor;
    FFontColorFocused: TColor;
    FFontColorDisabled: TColor;
    FMouseInControl: Boolean;
    FOnEnter: TNotifyEvent;
    FOnExit: TNotifyEvent;
    FOnMouseEnter: TNotifyEvent;
    FOnMouseLeave: TNotifyEvent;
    function IsFontStored: Boolean;
    procedure SetColorDefault(Value: TColor);
    procedure SetColorDisabled(Value: TColor);
    procedure SetColorFocused(Value: TColor);
    procedure SetEnabled(Value: Boolean);
    procedure SetFont(Value: TFont);
    procedure SetFontColorFocused(Value: TColor);
    procedure SetFontColorDisabled(Value: TColor);
    procedure ChangeColors;
    procedure FontChanged(Sender: TObject);
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
  protected
    procedure MouseEnter; virtual;
    procedure MouseLeave; virtual;
    procedure DoEnter; Override;
    procedure DoExit; Override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: char); override;
    property Color read FColorDefault write SetColorDefault default clWindow;
    property ColorDisabled: TColor read FColorDisabled write SetColorDisabled default clWindow;
    property ColorFocused: TColor read FColorFocused write SetColorFocused default clWindow;
    property Enabled read FEnabled write SetEnabled default True;
    property ExitOnReturn: Boolean read FExitOnReturn write FExitOnReturn default False;
    property Font read FFont write SetFont stored IsFontStored;
    property MouseInControl: Boolean read FMouseInControl;
    property About: TKNGAboutString read FAbout write FAbout;
    property FocusOnMouseEnter: Boolean read FFocusOnMouseEnter write FFocusOnMouseEnter;
    property FontColorFocused: TColor read FFontColorFocused write SetFontColorFocused default clBlack;
    property FontColorDisabled: TColor read FFontColorDisabled write SetFontColorDisabled default clBlack;
    property OnEnter read FOnEnter write FOnEnter;
    property OnExit read FOnExit write FOnExit;
    property OnMouseEnter: TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave: TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

  TKNGEdit = class(TKNGCustomEdit)
  published
    { Properties }
    { Inherited from TCustomEdit }
    property AutoSelect;
    property AutoSize;
    property BorderStyle;
    property CharCase;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property ImeMode;
    property ImeName;
    property MaxLength;
    property OEMConvert;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PasswordChar;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Text;
    property Visible;
    { Inherited from TKNGCustomEdit }
    property About;
    property ColorDisabled;
    property ColorFocused;
    property ExitOnReturn;
    property FocusOnMouseEnter;
    property FontColorFocused;
    property FontColorDisabled;
    { Delphi 4 Properties }
    {$IFDEF KNG_D4}
    property Anchors;
    property BiDiMode;
    property Constraints;
    property DragKind;
    property ParentBiDiMode;
    {$ENDIF}
    { Events }
    { Inherited from TCustomEdit}
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
    { Inherited from TKNGCustomEdit }
    property OnMouseEnter;
    property OnMouseLeave;
    { Delphi 4 Events }
    {$IFDEF KNG_D4}
    property OnEndDock;
    property OnStartDock;
    {$ENDIF}
  end;

  TKNGCustomPanel = class(TCustomPanel)
  private
    FAbout          : TKNGAboutString;
    FBevel          : TKNGBevel;
    FMouseInControl : Boolean;
    FOnMouseEnter   : TNotifyEvent;
    FOnMouseLeave   : TNotifyEvent;

    Function GetClientHeight : Integer;
    function GetClientWidth  : Integer;
    procedure SetBevel(Value: TKNGBevel);
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
  protected
    procedure AlignControls(AControl: TControl; var Rect: TRect); override;
    procedure BevelChanged(Sender: TObject); virtual;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure MouseEnter; virtual;
    procedure MouseLeave; virtual;
    procedure Paint; override;
    property About: TKNGAboutString read FAbout write FAbout;
    property Bevel: TKNGBevel read FBevel write SetBevel;
    property ClientHeight: Integer read GetClientHeight;
    property ClientWidth: Integer read GetClientWidth;
    property OnMouseEnter: TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave: TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

  TKNGPanel = class(TKNGCustomPanel)
  published
    property About;
    property Align;
    property Alignment;
    property Bevel;
    property DragCursor;
    property DragMode;
    property Enabled;
    property FullRepaint;
    property Caption;
    property Color;
    property Ctl3D;
    property Font;
    property Locked;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDrag;
  end;

implementation

uses Forms, ShellApi, SysUtils, KNGgUtil, Dialogs;

{ TKNGBevel }

procedure TKNGBevel.Assign(Source: TPersistent);
begin
  if Source is TKNGBevel then
  begin
    FBevelInner := TKNGBevel(Source).BevelInner;
    FBevelOuter := TKNGBevel(Source).BevelOuter;
    FBevelWidth := TKNGBevel(Source).BevelWidth;
    FBevelBorders := TKNGBevel(Source).BevelBorders;
    FBorderStyle := TKNGBevel(Source).BorderStyle;
    FBorderWidth := TKNGBevel(Source).BorderWidth;
    FHighLightColor := TKNGBevel(Source).HighLightColor;
    FShadowColor := TKNGBevel(Source).ShadowColor;
    if Assigned(FOnChange) then FOnChange(Self);
  end
  else
    inherited Assign(Source);
end;

function TKNGBevel.ClientRect(ControlRect: TRect): TRect;
begin
  Result := ControlRect;
end;

constructor TKNGBevel.Create;
begin
  inherited Create;
  FBevelInner := bvNone;
  FBevelOuter := bvRaised;
  FBevelWidth := 1;
  FBevelBorders := [bsLeft, bsRight, bsTop, bsBottom];
  FBorderStyle := bsNone;
  FBorderWidth := 0;
  FHighLightColor := clBtnHighLight;
  FShadowColor := clBtnShadow;
  FOnChange := nil;
end;

destructor TKNGBevel.Destroy;
begin
  FOnChange := nil;
  inherited Destroy;
end;

function TKNGBevel.GetBevelSideSize : Integer;
begin
  Result := 0;

  { Add the BorderWidth }
  Inc(Result, FBorderWidth);

  { If both BevelStyles are bsNone then we can Exit }
  if (FBevelOuter=bvNone) and (FBevelInner=bvNone) then exit;

  { Now we should add the width of the Inner and Outer Bevel if they are not
    set to bvNone }
  if FBevelOuter <> bvNone then
    Inc(Result, FBevelWidth);
  if FBevelInner <> bvNone then
    Inc(Result, FBevelWidth);
end;

procedure TKNGBevel.PaintBevel(Canvas: TCanvas; var Rect: TRect; Color: TColor; Transparent: Boolean);
var
  TopColor, BottomColor : TColor;

  procedure AdjustColors(Bevel: TKNGBevelStyle);
  begin
    TopColor := FHighLightColor;
    if Bevel = bvLowered then TopColor := FShadowColor;
    BottomColor := FShadowColor;
    if Bevel = bvLowered then BottomColor := FHighLightColor;
  end;

begin
  if FBevelOuter <> bvNone then
  begin
    AdjustColors(FBevelOuter);
    DrawFrame3d(Canvas, Rect, TopColor, BottomColor, FBevelWidth, FBevelBorders, Color);
  end;

  if Transparent then
    InflateRect(Rect, - FBorderWidth, - FBorderWidth)
  else
    DrawFrame3d(Canvas, Rect, Color, Color, FBorderWidth, FBevelBorders, Color);
  if FBevelInner <> bvNone then
  begin
    AdjustColors(FBevelInner);
    DrawFrame3d(Canvas, Rect, TopColor, BottomColor, FBevelWidth, FBevelBorders, Color);
  end;
  InflateRect(Rect, -1, -1);
end;

procedure TKNGBevel.SetBevelBorders(Value: TKNGBorderSides);
begin
  if FBevelBorders <> Value then
  begin
    FBevelBorders := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TKNGBevel.SetBevelInner(Value: TKNGBevelStyle);
begin
  if FBevelInner <> Value then
  begin
    FBevelInner := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TKNGBevel.SetBevelOuter(Value: TKNGBevelStyle);
begin
  if FBevelOuter <> Value then
  begin
    FBevelOuter := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TKNGBevel.SetBevelWidth(Value: TBevelWidth);
begin
  if FBevelWidth <> Value then
  begin
    FBevelWidth := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TKNGBevel.SetBorderStyle(Value: TKNGBorderStyle);
begin
  if FBorderStyle <> Value then
  begin
    FBorderStyle := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TKNGBevel.SetBorderWidth(Value: TBorderWidth);
begin
  if FBorderWidth <> Value then
  begin
    FBorderWidth := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TKNGBevel.SetHighLightColor(Value: TColor);
begin
  if FHighLightColor <> Value then
  begin
    FHighLightColor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

procedure TKNGBevel.SetShadowColor(Value: TColor);
begin
  if FShadowColor <> Value then
  begin
    FShadowColor := Value;
    if Assigned(FOnChange) then FOnChange(Self);
  end;
end;

{ TKNGGraphicControl }

procedure TKNGGraphicControl.CMMouseEnter(var Message: TMessage);
begin
  inherited;
  if not FMouseInControl and Enabled then
  begin
    FMouseInControl := True;
    MouseEnter;
  end;
end;

procedure TKNGGraphicControl.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  if FMouseInControl and Enabled then
  begin
    FMouseInControl := False;
    MouseLeave;
  end;
end;

procedure TKNGGraphicControl.MouseEnter;
begin
  if Assigned(FOnMouseEnter) then FOnMouseEnter(self);
end;

procedure TKNGGraphicControl.MouseLeave;
begin
  if Assigned(FOnMouseLeave) then FOnMouseLeave(self);
end;

{ TKNGCustomLabel }

procedure TKNGCustomLabel.AdjustBounds;
const
  WordWraps: array[Boolean] of Word = (0, DT_WORDBREAK);
var
  DC: HDC;
  X: Integer;
  Rect: TRect;
begin
  if not (csReading in ComponentState) and FAutoSize then
  begin
    Rect := ClientRect;
    DC := GetDC(0);
    Canvas.Handle := DC;
    DoDrawText(Rect, (DT_EXPANDTABS or DT_CALCRECT) or WordWraps[FWordWrap]);
    Inc(Rect.Right,  2 * (FBevel.BevelSideSize + 1));
    Inc(Rect.Bottom, 2 * (FBevel.BevelSideSize + 1));
    SetBounds(Left, Top, Rect.Right, Rect.Bottom);
    Canvas.Handle := 0;
    ReleaseDC(0, DC);
    X := Left;
    if FAlignment = taRightJustify then Inc(X, Width - Rect.Right);
    SetBounds(X, Top, Rect.Right, Rect.Bottom);
  end;
end;

constructor TKNGCustomLabel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csOpaque, csReplicatable];
  FBevel := TKNGBevel.Create;
  FBevel.BevelOuter := bvNone;
  FBevel.OnChange := BevelChanged;
  Width := 65;
  Height := 17;
  FAutoSize := True;
  FShowAccelChar := True;
end;

destructor TKNGCustomLabel.Destroy;
begin
  FBevel.OnChange := nil;
  FBevel.Free;
  FBevel := nil;
  inherited Destroy;
end;

procedure TKNGCustomLabel.BevelChanged(Sender: TObject);
begin
  AdjustBounds;
  Paint;
end;

procedure TKNGCustomLabel.CMDialogChar(var Message: TCMDialogChar);
begin
  if  (FFocusControl <> nil) and Enabled
  and ShowAccelChar and IsAccel(Message.CharCode, Caption) then
  begin
    with FFocusControl do
    begin
      if CanFocus then
      begin
        SetFocus;
        Message.Result := 1;
      end;
    end;
  end;
end;

procedure TKNGCustomLabel.CMFontChanged(var Message: TMessage);
begin
  inherited;
  AdjustBounds;
end;

procedure TKNGCustomLabel.CMTextChanged(var Message: TMessage);
begin
  Invalidate;
  AdjustBounds;
end;

procedure TKNGCustomLabel.DoDrawText(var Rect: TRect; Flags: Word);
var
  Text: string;
begin
  Text := GetLabelText;
  if (Flags and DT_CALCRECT <> 0) and ((Text = '') or FShowAccelChar and
    (Text[1] = '&') and (Text[2] = #0)) then Text := Text + ' ';
  if not FShowAccelChar then Flags := Flags or DT_NOPREFIX;
  Canvas.Font := Font;
  if not Enabled then
  begin
    OffsetRect(Rect, 1, 1);
    Canvas.Font.Color := clBtnHighlight;
    DrawText(Canvas.Handle, PChar(Text), Length(Text), Rect, Flags);
    OffsetRect(Rect, -1, -1);
    Canvas.Font.Color := clBtnShadow;
    DrawText(Canvas.Handle, PChar(Text), Length(Text), Rect, Flags);
  end
  else
    DrawText(Canvas.Handle, PChar(Text), Length(Text), Rect, Flags);
end;

function TKNGCustomLabel.GetLabelText: string;
begin
  Result := Caption;
end;

function TKNGCustomLabel.GetTransparent: Boolean;
begin
  Result := not (csOpaque in ControlStyle);
end;

procedure TKNGCustomLabel.Loaded;
begin
  inherited Loaded;
  AdjustBounds;
end;

procedure TKNGCustomLabel.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FFocusControl) then
    FFocusControl := nil;
end;

procedure TKNGCustomLabel.Paint;
const
  Alignments: array[TAlignment] of Word = (DT_LEFT, DT_RIGHT, DT_CENTER);
  WordWraps: array[Boolean] of Word = (0, DT_WORDBREAK);
var
  Rect, CalcRect: TRect;
  DrawStyle: Integer;
begin
  with Canvas do
  begin
    if not Transparent then
    begin
      Brush.Color := Self.Color;
      Brush.Style := bsSolid;
      FillRect(ClientRect);
    end;

    Rect := ClientRect;
    FBevel.PaintBevel(Canvas, Rect, Color, Transparent);

    Brush.Style := bsClear;
    DrawStyle := DT_EXPANDTABS or WordWraps[FWordWrap] or Alignments[FAlignment];
    { Calculate vertical layout }
    if FLayout <> tlTop then
    begin
      CalcRect := Rect;
      DoDrawText(CalcRect, DrawStyle or DT_CALCRECT);
      if FLayout = tlBottom then OffsetRect(Rect, 0, Height - CalcRect.Bottom)
      else OffsetRect(Rect, 0, (Height - CalcRect.Bottom - (FBevel.BevelSideSize + 1)) div 2);
    end;
    DoDrawText(Rect, DrawStyle);
  end;
end;

procedure TKNGCustomLabel.SetAlignment(Value: TAlignment);
begin
  if FAlignment <> Value then
  begin
    FAlignment := Value;
    Invalidate;
  end;
end;

procedure TKNGCustomLabel.SetAutoSize(Value: Boolean);
begin
  if FAutoSize <> Value then
  begin
    FAutoSize := Value;
    AdjustBounds;
  end;
end;

procedure TKNGCustomLabel.SetBevel(Value: TKNGBevel);
begin
  if Value <> FBevel then
  begin
    FBevel.Assign(Value);
  end;
end;

procedure TKNGCustomLabel.SetFocusControl(Value: TWinControl);
begin
  FFocusControl := Value;
  if Value <> nil then Value.FreeNotification(Self);
end;

procedure TKNGCustomLabel.SetLayout(Value: TTextLayout);
begin
  if FLayout <> Value then
  begin
    FLayout := Value;
    Invalidate;
  end;
end;

procedure TKNGCustomLabel.SetShowAccelChar(Value: Boolean);
begin
  if FShowAccelChar <> Value then
  begin
    FShowAccelChar := Value;
    Invalidate;
  end;
end;

procedure TKNGCustomLabel.SetTransparent(Value: Boolean);
begin
  if Transparent <> Value then
  begin
    if Value then
      ControlStyle := ControlStyle - [csOpaque] else
      ControlStyle := ControlStyle + [csOpaque];
    Invalidate;
  end;
end;

procedure TKNGCustomLabel.SetWordWrap(Value: Boolean);
begin
  if FWordWrap <> Value then
  begin
    FWordWrap := Value;
    AdjustBounds;
    Invalidate;
  end;
end;

{ TKNGURLLabel }

constructor TKNGURLLabel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Height := 30;
  Font.Color := clBlue;
  Font.Style := [fsUnderLine];
end;

destructor TKNGURLLabel.Destroy;
begin
  inherited Destroy;
end;

procedure TKNGURLLabel.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  URLString: string;
begin
  inherited MouseDown(Button, Shift, X, Y);
  if Caption <> '' then
  begin
    URLString := '';
    if FAddURLKind then
    begin
      case FURLKIND of
        ukHTTP   : URLString := URLString + 'http://';
        ukMailTo : URLString := URLString + 'mailto:';
      end;
    end;
    URLString := URLString + Caption;
    ShellExecute(ValidParentForm(Self).Handle,'Open', PChar(URLString),
                 nil, nil, SW_SHOWNORMAL);
  end;
end;

procedure TKNGURLLabel.MouseEnter;
begin
  if Assigned(FOnMouseEnter) then FOnMouseEnter(self);
  cursor := crHandPoint;
end;

procedure TKNGURLLabel.MouseLeave;
begin
  cursor := crDefault;
  if Assigned(FOnMouseLeave) then FOnMouseLeave(self);
end;

procedure TKNGURLLabel.SetAddURLKind(Value: Boolean);
begin
  if Value <> FAddURLKind then
    FAddURLKind := Value
end;

procedure TKNGURLLabel.SetURLKind(Value: TKNGURLKind);
begin
  if Value <> FURLKind then
    FURLKind := Value;
end;

{ TKNGCustomEdit }

procedure TKNGCustomEdit.ChangeColors;
begin
  if FEnabled then
  begin
    if Focused then
    begin
      inherited Color := FColorFocused;
      inherited Font.Color := FFontColorFocused;
    end
    else
    begin
      inherited Color := FColorDefault;
      inherited Font.Color := FFontColorDefault;
    end;
  end
  else
  begin
    inherited Color := FColorDisabled;
    inherited Font.Color := FFontColorDisabled;
  end;
  Invalidate;
end;

procedure TKNGCustomEdit.CMEnter(var Message: TCMEnter);
begin
  inherited;
  DoEnter;
end;

procedure TKNGCustomEdit.CMExit(var Message: TCMExit);
begin
  DoEnter;
end;

procedure TKNGCustomEdit.CMMouseEnter(var Message: TMessage);
begin
  if Enabled then
  begin
    inherited;
    if not FMouseInControl then
    begin
      FMouseInControl := True;
      MouseEnter;
    end;
  end
end;

procedure TKNGCustomEdit.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  if FMouseInControl and Enabled then
  begin
    FMouseInControl := False;
    MouseLeave;
  end;
end;

constructor TKNGCustomEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FFont := TFont.Create;
  FFont.OnChange := FontChanged;

  FFontColorDefault := FFont.Color;
  FFontColorDisabled := clBlack;
  FFontColorFocused := clBlack;
  FColorDefault := clWindow;
  FColorDisabled := clWindow;
  FColorFocused := clWindow;
  FFocusOnMouseEnter := False;
  FEnabled := True;
  FExitOnReturn := False;
  FFont.OnChange := FontChanged;
end;

destructor TKNGCustomEdit.Destroy;
begin
  Font.OnChange := nil;

  FFont.Free;
  FFont := nil;

  inherited Destroy;
end;

procedure TKNGCustomEdit.DoEnter;
begin
  if Assigned(FOnEnter) then FOnEnter(Self);
  ChangeColors;
end;

procedure TKNGCustomEdit.DoExit;
begin
  ChangeColors;
  if Assigned(FOnExit) then FOnExit(Self);
end;

procedure TKNGCustomEdit.FontChanged(Sender: TObject);
begin
  inherited ParentFont := False;
  if Font.Color <> FFontColorDefault then
    if (Focused = False) and Enabled then
    begin
      FFontColorDefault := Font.Color;
      ChangeColors;
    end;
  Perform(CM_FONTCHANGED, 0, 0);
end;

function TKNGCustomEdit.IsFontStored: Boolean;
begin
  Result := not ParentFont and not DesktopFont;
end;

procedure TKNGCustomEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  if Key = VK_RETURN then
    if (FExitOnReturn) then
      PostMessage(Handle, WM_KeyDown, VK_TAB, 0)
    else
    begin
      Key := 0;
      beep;
    end
  else
    inherited KeyDown(Key, Shift);
end;

procedure TKNGCustomEdit.KeyPress(var Key: char);
begin
  if (FExitOnReturn) and (Ord(Key) = VK_Return) then
    Key := char(VK_TAB);
  inherited KeyPress(Key);
end;

procedure TKNGCustomEdit.MouseEnter;
begin
  if Assigned(FOnMouseEnter) then FOnMouseEnter(self);
  if FFocusOnMouseEnter then
    if CanFocus and FEnabled then
      SetFocus;
end;

procedure TKNGCustomEdit.MouseLeave;
begin
  if Assigned(FOnMouseLeave) then FOnMouseLeave(self);
end;

procedure TKNGCustomEdit.SetColorDefault(Value: TColor);
begin
  if FColorDefault <> Value then
  begin
    FColorDefault := Value;
    ChangeColors;
  end;
end;

procedure TKNGCustomEdit.SetColorDisabled(Value: TColor);
begin
  if FColorDisabled <> Value then
  begin
    FColorDisabled := Value;
    ChangeColors;
  end;
end;

procedure TKNGCustomEdit.SetColorFocused(Value: TColor);
begin
  if FColorFocused <> Value then
  begin
    FColorFocused := Value;
    ChangeColors;
  end;
end;

procedure TKNGCustomEdit.SetEnabled(Value: Boolean);
begin
  if FEnabled <> Value then
  begin
    inherited Enabled := Value;
    FEnabled := Value;
    ChangeColors;
  end;
end;

procedure TKNGCustomEdit.SetFont(Value: TFont);
begin
  FFont.Assign(Value);
end;

procedure TKNGCustomEdit.SetFontColorDisabled(Value: TColor);
begin
  if FFontColorDisabled <> Value then
  begin
    FFontColorDisabled := Value;
    ChangeColors;
  end;
end;

procedure TKNGCustomEdit.SetFontColorFocused(Value: TColor);
begin
  if FFontColorFocused <> Value then
  begin
    FFontColorFocused := Value;
    ChangeColors;
  end;
end;

{ TKNGCustomPanel }

procedure TKNGCustomPanel.AlignControls(AControl: TControl;
  var Rect: TRect);
var
  BevelSize: Integer;
begin
  BevelSize := FBevel.BorderWidth;
  if FBevel.BevelOuter <> bvNone then Inc(BevelSize, FBevel.BevelWidth);
  if FBevel.BevelInner <> bvNone then Inc(BevelSize, FBevel.BevelWidth);
  InflateRect(Rect, -BevelSize, -BevelSize);
  inherited AlignControls(AControl, Rect);
end;

procedure TKNGCustomPanel.BevelChanged(Sender: TObject);
begin
  RecreateWnd;
  ReAlign;
  Invalidate;
end;

procedure TKNGCustomPanel.CMMouseEnter(var Message: TMessage);
begin
  inherited;
  if not FMouseInControl and Enabled then
  begin
    FMouseInControl := True;
    MouseEnter;
  end;
end;

procedure TKNGCustomPanel.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  if FMouseInControl and Enabled then
  begin
    FMouseInControl := False;
    MouseLeave;
  end;
end;

constructor TKNGCustomPanel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FBevel := TKNGBevel.Create;
  FBevel.OnChange := BevelChanged;
  Width := 185;
  Height := 41;
  Alignment := taCenter;
  Color := clBtnFace;
  FullRepaint := True;
end;

procedure TKNGCustomPanel.CreateParams(var Params: TCreateParams);
const
  BorderStyles: array[TKNGBorderStyle] of DWORD = (0, WS_BORDER);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style := Style or BorderStyles[FBevel.BorderStyle];
    if NewStyleControls and Ctl3D and (FBevel.BorderStyle = bsSingle) then
    begin
      Style := Style and not WS_BORDER;
      ExStyle := ExStyle or WS_EX_CLIENTEDGE;
    end;
    WindowClass.style := WindowClass.style and not (CS_HREDRAW or CS_VREDRAW);
  end;
end;

destructor TKNGCustomPanel.Destroy;
begin
  FBevel.OnChange := nil;
  FBevel.Free;
  FBevel := nil;
  inherited Destroy;
end;

function TKNGCustomPanel.GetClientHeight : Integer ;
begin
  Result := Height - 2 * (FBevel.BevelSideSize + 1);
end;

function TKNGCustomPanel.GetClientWidth : Integer ;
begin
  Result := Width - 2 * (FBevel.BevelSideSize + 1);
end;

procedure TKNGCustomPanel.MouseEnter;
begin
  if Assigned(FOnMouseEnter) then FOnMouseEnter(self);
end;

procedure TKNGCustomPanel.MouseLeave;
begin
  if Assigned(FOnMouseLeave) then FOnMouseLeave(self);
end;

procedure TKNGCustomPanel.Paint;
var
  Rect : TRect;
  FontHeight : Integer;
const
  Alignments : array[TAlignment] of Word = (DT_LEFT, DT_RIGHT, DT_CENTER);
begin
  Rect := GetClientRect;

  { Call the TKNGBevel's PaintBevel method to draw the Bevel Borders }
  FBevel.PaintBevel(Canvas, Rect, Color, False);

  with Canvas do
  begin
    { Fill the panel's background }
    Brush.Color := Color;
    FillRect(Rect);

    { Now draw the Caption's Text on the Panel }
    Brush.Style := bsClear;
    Font := Self.Font;
    FontHeight := TextHeight('W');
    with Rect do
    begin
      Top := ((Bottom + Top) - FontHeight) div 2;
      Bottom := Top + FontHeight;
    end;
    DrawText(Handle, PChar(Caption),  -1, Rect, (DT_EXPANDTABS or
             DT_VCENTER) or Alignments[Alignment]);
  end;
end;

procedure TKNGCustomPanel.SetBevel(Value: TKNGBevel);
begin
  if Value <> FBevel then
  begin
    FBevel.Assign(Value);
  end;
end;

end.

